﻿using Unit4.CollectionsLib;

/// <summary>
/// פתרון למשימה מספר 2 מפרק 11 מפה - ייצוג המפה בעזרת עץ חוליות בינרי המאורגן כעץ-חיפוש-בינרי
/// ==========================================================================================
/// </summary>
/// <author> צוות מדעי המחשב, המרכז להוראת המדעים, האוניברסיטה העברית, ירושלים </author>
/// <version> 28.11.2007 </version>
/// <typeparam name="V">טיפוס הערכים הקשורים למפתחות</typeparam>

public class Map<V>
{
    private BinTreeNode<Pair<string, V>> bst;
    private int numOfKeys;


    public Map()
    {
        this.bst = null;
        this.numOfKeys = 0;
    }


    public bool ContainsKey(string key)
    {
        return SearchKey(key) != null;
    }

    // הנחה: המפתח קיים במפה
    public V GetValue(string key)
    {
        BinTreeNode<Pair<string, V>> pos = SearchKey(key);

        return pos.GetInfo().GetValue();
    }


    public void Insert(string key, V value)
    {
        BinTreeNode<Pair<string, V>> pos = SearchKey(key);

        if (pos != null)
            pos.GetInfo().SetValue(value);
        else
        {
            this.numOfKeys++;
            pos = this.bst;
            bool inTree = false;
            BinTreeNode<Pair<string, V>> node = new BinTreeNode<Pair<string, V>>(new Pair<string, V>(key, value));

            if (this.bst == null)
            {
                this.bst = node;
                inTree = true;
            }
            while (!inTree)
            {
                if (key.CompareTo(pos.GetInfo().GetKey()) < 0)
                {
                    if (pos.GetLeft() == null)
                    {
                        pos.SetLeft(node);
                        inTree = true;
                    }
                    else
                        pos = pos.GetLeft();
                }
                else
                {
                    if (pos.GetRight() == null)
                    {
                        pos.SetRight(node);
                        inTree = true;
                    }
                    else
                        pos = pos.GetRight();
                }
            }
        }
    }


    // הנחה: המפתח קיים במפה
    public V Remove(string key)
    {
        BinTreeNode<Pair<string, V>> pos = SearchKey(key);
        BinTreeNode<Pair<string, V>> next = null;
        BinTreeNode<Pair<string, V>> parent = GetParent(pos);

        this.numOfKeys--;
        if ((pos.GetLeft() == null) || (pos.GetRight() == null))
        {
            if (pos.GetLeft() != null)
            {
                next = pos.GetLeft();
                pos.SetLeft(null);
            }
            else
            {
                next = pos.GetRight();
                pos.SetRight(null);
            }
            if (parent == null)
                this.bst = next;
            else
                if (parent.GetLeft() == pos)
                    parent.SetLeft(next);
                else
                    parent.SetRight(next);
        }
        else
        {
            next = pos.GetRight();
            while (next.GetLeft() != null)
                next = next.GetLeft();
            pos.SetInfo(next.GetInfo());
            if (pos.GetRight() != next)
            {
                parent = GetParent(next);
                parent.SetLeft(next.GetRight());
            }
            else
                pos.SetRight(next.GetRight());
            next.SetRight(null);
        }

        return pos.GetInfo().GetValue();
    }


    public string[] GetAllKeys()
    {
        int i = 0;
        BinTreeNode<Pair<string, V>> tmp = this.bst;
        string[] keys = new string[this.numOfKeys];
        Stack<BinTreeNode<Pair<string, V>>> stk = new Stack<BinTreeNode<Pair<string, V>>>();

        do
        {
            while (tmp != null)
            {
                stk.Push(tmp);
                tmp = tmp.GetLeft();
            }
            if (!stk.IsEmpty())
            {
                tmp = stk.Pop();
                keys[i] = tmp.GetInfo().GetKey();
                i++;
                tmp = tmp.GetRight();
            }
        } while (!stk.IsEmpty() || tmp != null);

        return keys;
    }


    public override string ToString()
    {
        int i;
        string str = "[";
        string[] keys = this.GetAllKeys();

        for (i = 0; i < keys.Length - 1; i++)
            str = str + keys[i] + ":" + this.GetValue(keys[i]).ToString() + ", ";
        if (keys.Length > 0)
            str = str + keys[i] + ":" + this.GetValue(keys[i]).ToString();
        str = str + "]";

        return str;
    }

    // -------------------------------------------------------------
    // ------------------ private methods for help -----------------
    // -------------------------------------------------------------

    // הפעולה מחזירה הפניה לצומת בעץ שהפתח שלו מתקבל כפרמטר
    private BinTreeNode<Pair<string, V>> SearchKey(string key)
    {
        BinTreeNode<Pair<string, V>> p = this.bst;

        while (p != null)
        {
            if (key.CompareTo(p.GetInfo().GetKey()) == 0)
                return p;
            if (key.CompareTo(p.GetInfo().GetKey()) < 0)
                p = p.GetLeft();
            else
                p = p.GetRight();
        }
        return null;
    }

    // הפעולה מחזירה את ההורה של צומת נתון
    private BinTreeNode<Pair<string, V>> GetParent(BinTreeNode<Pair<string, V>> node)
    {
        BinTreeNode<Pair<string, V>> p = this.bst;

        while (p != null)
        {
            if ((node == p.GetLeft()) || (node == p.GetRight()))
                return (p);
            if (node.GetInfo().GetKey().CompareTo(p.GetInfo().GetKey()) >= 0)
                p = p.GetRight();
            else
                p = p.GetLeft();
        }
        return (null);
    }
}